/* ---------------------------------------------------------------------------------------------- *
 *          P R O J E C T :   S O L D E R I N G   O V E N                                         *
 * ---------------------------------------------------------------------------------------------- *
 *
 * main.c
 *
 *  Created: 26/05/2014 1:52:13 PM
 * Modified: 05/05/2015 1:23:00 PM - Reduce Tmin to 100C
 *   Author: David
 * 
 * CPU: Atmel AtMega8-16AU
 * Clock source: internal, calibrated RC oscillator (1MHz)
 *
 * IO Environment:
 * PB0 \BTN0	PC0  BRD_TEMP	PD0  RXD (RS485)
 * PB1 \EN		PC1  OVEN_TEMP	PD1	 TXD    "
 * PB2  LD		PC2  --			PD2	 DE     "
 * PB3  MOSI	PC3  --			PD3 \LED3
 * PB4  MISO	PC4  --			PD4 \BTN3
 * PB5  SCK		PC5  --			PD5 \LED1
 * PB6 \LED2	PC6 \RESET		PD6 \BTN1
 * PB7 \BTN2	PC7  --			PD7 \LED0
 *
 * SPI Output (16 bits):
 * 15  DGT2  Byte 0 MSB: first out
 * 14  DGT1
 * 13  DGT0
 * 12  CNTRL_FAN
 * 11  HTR2
 * 10  AC_FAN
 *  9  HTR1
 *  8  --
 *  7  SEGG
 *  6  SEGE
 *  5  SEGC
 *  4  SEGA
 *  3  SEGB
 *  2  SEGD
 *  1  SEGF
 *  0  DP    Byte 1 LSB: last out
 *
 * Strategy:
 * The oven heats under power at about 1.5C / sec.
 * This is so slow that there is no need for a PID controller.
 * 
 * Sequence:
 * 1. Power on, let temp. rise to T1. Set Timer = E1.
 * 2. Power on, let temp. rise to T2 - Offset.
 * 3. while(Timer < E2) power off.
 * 4. Power on, let temp. rise to T3 - Offset.
 * 5. while(Timer < E3) power off.
 * 6. Power on, let temp. rise to T4 - Offset.
 * 7. Power off & cool down.
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdbool.h>
#include "main.h"

S_CONFIG Config =
{
	{									// Profile for leaded solder
	//   E0  T0    E1 T1     E2  T2     E3  T3     E4  T4     E5  T5
		{ 0, 25}, {30,100}, {120,150}, {150,183}, {210,235}, {240,183} 
	},
	{									// Profile for unleaded solder
		{ 0, 25}, {90,150}, {180,175}, {210,217}, {240,249}, {270,217}
	},
	   60,								// WarmController_u16
	   40,								// CoolController_u16
	  100,								// HotController_u16
	   65,								// OvenFanOnTemp_u16
	   55,								// OvenFanOffTemp_u16
	   10,								// TempOvershoot1_u16
	   10,								// TempOvershoot2_u16
	15625,								// T1Adivide_u16
	BV(CS11),							// T1Aprescale_u8
	BV(ADPS2) | BV(ADPS1) | BV(ADPS0),	// AdcPrescale_u8
	BV(SPR1)  | BV(SPR0)				// SpiPrescale_u8
};

typedef enum { OVEN_E, CNTLR_E, PROFILE_E, TIME_E } DISPSTATE_E;
	
/* ---------------------------------------------------------------------------------------------- *
 * Global variables                                                                               *
 * ---------------------------------------------------------------------------------------------- */
uint16_t OvenTemp_u16			= 0;			// Celsius, values updated transparently
uint16_t ControllerTemp_u16		= 0;

bool     timerflag_b			= false;		// Run the state machine
bool     Led_ab[NUMBUTTONS]		= { false };	// Button LEDs
uint8_t  CurProfile_u8			= 0;			// Current profile selection
	
const struct
{
	S_SPOINT *Profile_as;
	char     *Label_ac;
} Mode[] = { {Config.Lead_s, " Pb"}, {Config.UnLead_s, "-Pb"} };
	
#define NPROFILES (sizeof(Mode)/sizeof(Mode[0]))

uint16_t ElapsedSecs_u16 = 0;

/* ---------------------------------------------------------------------------------------------- *
 * Timer 1A: main sequencing & button de-bouncing                                                 *
 * System clock = 1MHz, final clock reqd. = 8Hz. So divisor = 125000                              *
 * Prescaler = 8, Timer = 15625 (NB Clock tolerance = 3%)                                         *
 * ---------------------------------------------------------------------------------------------- */
// Start the timer, driving the tick interrupt
static void TimerStartup(void)
{
	TCCR1A = 0;
	TCCR1B = BV(WGM12) | Config.T1Aprescale_u8;		// CTC mode, prescaler = 8
    OCR1A  = Config.T1Adivide_u16;					// Scale to 8Hz
    TIMSK |= BV(OCIE1A);							// Interrupt on
}

// The timer-tick function: main update period = 8Hz
SIGNAL(TIMER1_COMPA_vect)
{
	timerflag_b = true;				// Do the work at base level
	BlinkGate_b = !BlinkGate_b;		// Time blinked 7-segts.
}

/* ---------------------------------------------------------------------------------------------- *
 * Set the CPU pins as inputs & outputs, as reqd.                                                 *
 * ---------------------------------------------------------------------------------------------- */
static void IoLines(void)
{
	// Disable outputs at start: those set here are active-low
	PORTB = LD2_ | EN_;
	PORTD = LD0_ | LD1_ | LD3_;
	// Set pin output-drivers
	DDRB  = LD2_ | SCK  | MOSI | LD | EN_;
	DDRC  = 0;					// Used only for ADC (inputs)
	DDRD  = LD0_ | LD1_ | LD3_ | DE | TXD;
}

/* ---------------------------------------------------------------------------------------------- *
 * Panel LEDs                                                                                     *
 * ---------------------------------------------------------------------------------------------- */
static void LedUpdate(void)
{	
	if(Led_ab[0])PORTD &= ~LD0_; else PORTD |= LD0_;
	if(Led_ab[1])PORTD &= ~LD1_; else PORTD |= LD1_;
	if(Led_ab[2])PORTB &= ~LD2_; else PORTB |= LD2_;
	if(Led_ab[3])PORTD &= ~LD3_; else PORTD |= LD3_;
}
	
static void SetLed(uint8_t pos_u8, bool on_b)
{
	if (pos_u8 < NUMBUTTONS)
	{
		Led_ab[pos_u8] = on_b;
	}
}

/* ---------------------------------------------------------------------------------------------- *
 * Controller PCB temperature: DC fan control                                                     *
 * ---------------------------------------------------------------------------------------------- */
static bool ControllerTemp(void)
{
	if (ControllerTemp_u16 > Config.HotController_u16)
	{
		return false;
	} 
	else
	{
		if (ControllerTemp_u16 > Config.WarmController_u16)
		{
			SpiSetPin(SPI_DCFAN, true);
		}
	
		if (ControllerTemp_u16 < Config.CoolController_u16)
		{
			SpiSetPin(SPI_DCFAN, false);
		}
		return true;
	}
}

/* ---------------------------------------------------------------------------------------------- *
 * Display data as appropriate                                                                    *
 * ---------------------------------------------------------------------------------------------- */
static void DoDisplay(DISPSTATE_E disp_e)
{
	switch (disp_e)
	{
	case OVEN_E:
		putnum(OvenTemp_u16);
		break;
	case CNTLR_E:
		putnum(ControllerTemp_u16);
		putchr('C', false, false, 0);
		break;
	case PROFILE_E:
		putstr(Mode[CurProfile_u8].Label_ac, false);
		break;
	case TIME_E:
		putnum(ElapsedSecs_u16 / 6);	// Minutes * 10
		setdecimal(1, true);
		break;
	}
}

/* ---------------------------------------------------------------------------------------------- *
 * AC fan control with hysteresis                                                                 *
 * ---------------------------------------------------------------------------------------------- */
static void OvenFanControl(void)
{
	if (OvenTemp_u16 > Config.OvenFanOnTemp_u16)
	{
		SpiSetPin(SPI_ACFAN, true);
	}

	if (OvenTemp_u16 < Config.OvenFanOffTemp_u16)
	{
		SpiSetPin(SPI_ACFAN, false);
	}

}

/* ---------------------------------------------------------------------------------------------- *
 * Mainline                                                                                       *
 * ---------------------------------------------------------------------------------------------- */
int main(void)
{
	uint8_t  tbase_u8    = 0;
	bool     safe_b      = true;	// Set false if PCB overheats
	uint8_t  state_u8    = 0;
	uint16_t maxtemp_u16 = 0;
	
	DISPSTATE_E DispData_e = OVEN_E;
	
	IoLines();					// Set IO pin directions & default states
	AdcStartup();				// Initialize devices
	TimerStartup();
	SpiStartup();
    sei();						// Interrupts on
		
    while(1)
    {
		safe_b &= ControllerTemp();	// If PCB overheats, must power down to reset
		
		if(safe_b)					// PCB temperature OK?
		{							// PCB is OK: we can run
			OvenFanControl();
			
			if(ButtonHeld(0) && ButtonHeld(3))	// Buttons 0 & 3 together - 
			{									// Emergency stop!
				safe_b   = false;
				state_u8 = 0;	
				continue;						// Skip everything else				
			}
			
			if(timerflag_b)			// Fine timer interval: 8Hz
			{
				timerflag_b = false;
		
				ButtonUpdate();
				LedUpdate();
//				SetLed(0, (state_u8 > 0));

				
				if (ButtonPressed(0))		// Button[0] - cycle display
				{
					DispData_e = (DispData_e == TIME_E) ? OVEN_E : DispData_e+1;
				}
				DoDisplay(DispData_e);
				
				if (DispData_e == PROFILE_E)	// If profile is displayed, we can change it
				{
//					SetLed(1, true);
//					SetLed(2, true);
					if (ButtonPressed(1))			// Next profile
					{
						if((++CurProfile_u8) >= NPROFILES)
						{
							CurProfile_u8 = 0;
						}
					} 
					else if (ButtonPressed(2))		// Previous profile
					{
						if(CurProfile_u8 == 0)
						{
							CurProfile_u8 = NPROFILES;
						}
						CurProfile_u8--;
					}
				}
				else
				{
//					SetLed(1, false);
//					SetLed(2, false);
				}
																
				if (tbase_u8 >= 8)	// Coarse interval: 1Hz
				{
					tbase_u8 = 0;
					ElapsedSecs_u16++;
				} 
				else
				{
					tbase_u8++;
				}
				
				SpiSetPin(SPI_HTR2, SpiGetPin(SPI_HTR1));	// Don't change both together
				SetLed(0, (state_u8 & 1));
				SetLed(1, (state_u8 & 2));
				SetLed(2, (state_u8 & 4));
				SetLed(3, SpiGetPin(SPI_HTR1));
				
				switch(state_u8)			// Main state machine
				{	
				case 0:							// Idle
					if (ButtonPressed(3))					// Button 3: start cycle
					{
						state_u8 = 1;
					}
					break;
				case 1:							// Be sure we start from cold
					if(OvenTemp_u16 < Config.CoolController_u16)
					{
						SpiSetPin(SPI_HTR1, true);
						ElapsedSecs_u16 = 0;
						state_u8 = 2;
					}
					break;
				case 2:							// Power on, let temp. rise to T1. Set Timer = E1.
					if(OvenTemp_u16 >= Mode[CurProfile_u8].Profile_as[1].celsius_u16)
					{
						SpiSetPin(SPI_HTR1, false);
						ElapsedSecs_u16 = Mode[CurProfile_u8].Profile_as[1].secs_u16;	// Reset timer
						maxtemp_u16     = OvenTemp_u16;
						state_u8        = 3;
					}
					break;
				case 3:							// Power off, & coast to peak.
					if(OvenTemp_u16 > maxtemp_u16)
					{
						maxtemp_u16 = OvenTemp_u16;
					}
					else if(OvenTemp_u16 < (maxtemp_u16 - 2))
					{
						SpiSetPin(SPI_HTR1, true);
						state_u8 = 4;						
					}
					break;
				case 4:							// Power on, let temp. rise to T2 - Offset.
					SpiSetPin(SPI_HTR1, (OvenTemp_u16 < (Mode[CurProfile_u8].Profile_as[2].celsius_u16 - Config.TempOvershoot1_u16)));
					if(ElapsedSecs_u16 >= Mode[CurProfile_u8].Profile_as[2].secs_u16)
					{
						SpiSetPin(SPI_HTR1, true);
						state_u8 = 5;
					}
					break;
				case 5:							// while(Timer < E2) power off.
					if(OvenTemp_u16 >= (Mode[CurProfile_u8].Profile_as[4].celsius_u16 - Config.TempOvershoot1_u16))
					{
						SpiSetPin(SPI_HTR1, false);
						state_u8 = 6;
					}
					break;
				case 6:							// Power on, let temp. rise to T4 - Offset.
					if(OvenTemp_u16 >= Mode[CurProfile_u8].Profile_as[4].celsius_u16)
					{
						state_u8 = 0;
					}
					break;
				default:						// Scrambled: force recovery
					state_u8 = 0;
					break;
				}			
			}
		}
		else
		{									// Controller PCB too hot: must stop
			SpiSetPin(SPI_HTR1,  false);
			SpiSetPin(SPI_HTR2,  false);
			SpiSetPin(SPI_ACFAN, true);		// Keep the fans going
			SpiSetPin(SPI_DCFAN, true);
			putstr("Hot", true);
		}										
    }
}