/* ---------------------------------------------------------------------------------------------- *
 *
 *   Module      :   Proportional, Integral, Derivative Controller
 *   Copyright   :   (C)2003 Kermit Design
 *   Platform(s) :   Generic
 *   Limitations :
 *   $Workfile: PID.c $
 *   $Revision: 1.3 $
 *
 *-----------------------------------------------------------------------------------------------
 *
 *   Notes       :   REPLACE with Notes for module
 *
 * ---------------------------------------------------------------------------------------------- */

#define __PID_C

/*----- INCLUDES -------------------------------------------------------------------------------- */

// 2007.03.04 20:23  [YYYY.MM.DD] Last edited (CodeWright)
// 2014.06.05 10:06  DRB changed to generic format

#include <stdint.h>
#include <stdbool.h>
#include "pid.h"
#include <stdlib.h>

/* ---------------------------------------------------------------------------------------------- *
 * PidStartup()                                                                                   *
 * ------------                                                                                   *
 * Function:     Initialize Simple PID variables and pre-calculated numbers                       *
 * Return Value: none                                                                             *
 * Inputs:       (1) Pointer to PID structure to initialize                                       *
 * Remarks:      none                                                                             *
 * ---------------------------------------------------------------------------------------------- */
void PidStartup(S_PID *pid_ptr)
{
    int32_t temp_s32;

    // Work out max integral contribution
    temp_s32  = pid_ptr->calibration_i_max_s16;
    temp_s32 *= PID_GAIN_SCALING;                        // mul 16
    pid_ptr->calc_i_max_s32 = abs(temp_s32);

    // Work out max derivative contribution
    temp_s32  = pid_ptr->calibration_d_max_s16;
    temp_s32 *= PID_GAIN_SCALING;
    pid_ptr->calc_d_max_s32 = abs(temp_s32);

// Remaining variables set from calibration record

    // Zero-ed at start-up
//    pid_ptr->error_s16 = 0;
//    pid_ptr->d_sum_s32 = 0;
}

/* ---------------------------------------------------------------------------------------------- *
 * PidCalculate()                                                                                 *
 * ------------                                                                                   *
 * Function:     Calculate PID output                                                             *
 * Return Value: SINT2 - PID output                                                               *
 * Inputs:       (1) Pointer to PI structure to calculate                                         *
 * Remarks:      A simplified calculated PID function                                             *
 *               NB For oven, all outputs are unsigned: +ve only                                  *
 * ---------------------------------------------------------------------------------------------- */
int16_t PidCalculate(S_PID *simple_pid_ptr, int16_t fbval_s16)
{
    int32_t temp_s32;
    int32_t total_s32 = 0;
    int16_t prror_s16;

    simple_pid_ptr->feedback_value_s16 = fbval_s16;

    // Keep track of errors, previous and new
    prror_s16 = simple_pid_ptr->error_s16;

    // Determine error
    simple_pid_ptr->error_s16 = (simple_pid_ptr->setpoint_s16 - simple_pid_ptr->feedback_value_s16);

    /////////////////////////////////////////////////////////////////////////////
    //   C A L C U L A T E   P    T E R M  ( x  P I D _ G A I N _ S C A L I N G )
    /////////////////////////////////////////////////////////////////////////////
    total_s32 = (int32_t)simple_pid_ptr->error_s16 * simple_pid_ptr->calibration_p_gain_s16;

    /////////////////////////////////////////////////////////////////////////////
    //   C A L C U L A T E   I    T E R M  ( x  P I D _ G A I N _ S C A L I N G )
    /////////////////////////////////////////////////////////////////////////////
    simple_pid_ptr->i_sum_s32 += (int32_t)simple_pid_ptr->error_s16 * simple_pid_ptr->calibration_i_gain_s16;

    // Apply limits: unsigned +ve only
    if (simple_pid_ptr->i_sum_s32 > simple_pid_ptr->calc_i_max_s32)
    {
        simple_pid_ptr->i_sum_s32 = simple_pid_ptr->calc_i_max_s32;
    }
    else if (simple_pid_ptr->i_sum_s32 < 0)
    {
        simple_pid_ptr->i_sum_s32 = 0;
    }
    else
    {
        // no clipping needed
    }
    total_s32 += simple_pid_ptr->i_sum_s32;

    /////////////////////////////////////////////////////////////////////////////
    //   C A L C U L A T E   D    T E R M  ( x  P I D _ G A I N _ S C A L I N G )
    /////////////////////////////////////////////////////////////////////////////

    if(simple_pid_ptr->calibration_d_gain_s16 != 0)  // Any D term at all?
    {
        //goes positive if rising too fast
        temp_s32  = simple_pid_ptr->error_s16 - prror_s16;
        temp_s32 *= (int32_t)(simple_pid_ptr->calibration_d_gain_s16);
     
        if (temp_s32 > simple_pid_ptr->calc_d_max_s32)
        {
            temp_s32 = simple_pid_ptr->calc_d_max_s32;
        }
        else if (temp_s32 < -simple_pid_ptr->calc_d_max_s32)
        {
            temp_s32 = -simple_pid_ptr->calc_d_max_s32;
        }
     
        // Filter D TERM
        temp_s32 = (PID_D_FILTER - 1) * simple_pid_ptr->d_sum_s32 + temp_s32;
        simple_pid_ptr->d_sum_s32 = temp_s32 / PID_D_FILTER;
     
        total_s32 += simple_pid_ptr->d_sum_s32;
    }

    /////////////////////////////////////////////////////////////////////////////
    //   F I N A L   S U M   R E S U L T
    /////////////////////////////////////////////////////////////////////////////
    // Down-scale the result
    total_s32 += PID_GAIN_SCALING / 2;
    total_s32 /= PID_GAIN_SCALING;

    // Apply limits: 
    if (total_s32 > PID_MAX_OUTPUT)
    {
        total_s32 = PID_MAX_OUTPUT;
    }
    if (total_s32 < 0)
    {
        total_s32 = 0;
    }

    simple_pid_ptr->result_s16 = (int16_t)total_s32;

    return(simple_pid_ptr->result_s16);
}
